@@ -1,6 +1,7 @@
 /* POSIX extended headers for tar.
 
-   Copyright (C) 2003-2007, 2009-2010, 2012-2013 Free Software Foundation, Inc.
+   Copyright (C) 2003-2007, 2009-2010, 2012-2014, 2016-2017 Free
+   Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -754,6 +755,16 @@ xheader_decode (struct tar_stat_info *st)
 	continue;
     }
   run_override_list (keyword_override_list, st);
+
+  /* The archived (effective) file size is always set directly in tar header
+     field, possibly overridden by "size" extended header - in both cases,
+     result is now decoded in st->stat.st_size */
+  st->archive_file_size = st->stat.st_size;
+
+  /* The real file size (given by stat()) may be redefined for sparse
+     files in "GNU.sparse.realsize" extended header */
+  if (st->real_size_set)
+    st->stat.st_size = st->real_size;
 }
 
 static void
@@ -802,11 +813,11 @@ xheader_store (char const *keyword, struct tar_stat_info *st,
   t = locate_handler (keyword);
   if (!t || !t->coder)
     return;
-  if (xheader_keyword_deleted_p (keyword)
-      || xheader_keyword_override_p (keyword))
+  if (xheader_keyword_deleted_p (keyword))
     return;
   xheader_init (&st->xhdr);
-  t->coder (st, keyword, &st->xhdr, data);
+  if (!xheader_keyword_override_p (keyword))
+    t->coder (st, keyword, &st->xhdr, data);
 }
 
 void
@@ -1016,7 +1027,7 @@ xheader_string_end (struct xheader *xhdr, char const *keyword)
     }
   x_obstack_blank (xhdr, p);
   x_obstack_1grow (xhdr, '\n');
-  cp = obstack_next_free (xhdr->stk) - xhdr->string_length - p - 1;
+  cp = (char*) obstack_next_free (xhdr->stk) - xhdr->string_length - p - 1;
   memmove (cp + p, cp, xhdr->string_length);
   cp = stpcpy (cp, np);
   *cp++ = ' ';
@@ -1279,15 +1290,33 @@ path_coder (struct tar_stat_info const *st, char const *keyword,
   code_string (st->file_name, keyword, xhdr);
 }
 
+static void
+raw_path_decoder (struct tar_stat_info *st, char const *arg)
+{
+  decode_string (&st->orig_file_name, arg);
+  decode_string (&st->file_name, arg);
+  st->had_trailing_slash = strip_trailing_slashes (st->file_name);
+}
+
+
 static void
 path_decoder (struct tar_stat_info *st,
 	      char const *keyword __attribute__((unused)),
 	      char const *arg,
 	      size_t size __attribute__((unused)))
 {
-  decode_string (&st->orig_file_name, arg);
-  decode_string (&st->file_name, arg);
-  st->had_trailing_slash = strip_trailing_slashes (st->file_name);
+  if (! st->sparse_name_done)
+    raw_path_decoder (st, arg);
+}
+
+static void
+sparse_path_decoder (struct tar_stat_info *st,
+                     char const *keyword __attribute__((unused)),
+                     char const *arg,
+                     size_t size __attribute__((unused)))
+{
+  st->sparse_name_done = true;
+  raw_path_decoder (st, arg);
 }
 
 static void
@@ -1359,7 +1388,10 @@ sparse_size_decoder (struct tar_stat_info *st,
 {
   uintmax_t u;
   if (decode_num (&u, arg, TYPE_MAXIMUM (off_t), keyword))
-    st->stat.st_size = u;
+    {
+      st->real_size_set = true;
+      st->real_size = u;
+    }
 }
 
 static void
@@ -1442,13 +1474,13 @@ sparse_map_decoder (struct tar_stat_info *st,
 		    size_t size __attribute__((unused)))
 {
   int offset = 1;
+  struct sp_array e;
 
   st->sparse_map_avail = 0;
   while (1)
     {
       intmax_t u;
       char *delim;
-      struct sp_array e;
 
       if (!ISDIGIT (*arg))
 	{
@@ -1716,7 +1748,7 @@ struct xhdr_tab const xhdr_tab[] = {
   { "uname",    uname_coder,    uname_decoder,    0, false },
 
   /* Sparse file handling */
-  { "GNU.sparse.name",       path_coder, path_decoder,
+  { "GNU.sparse.name",       path_coder, sparse_path_decoder,
     XHDR_PROTECTED, false },
   { "GNU.sparse.major",      sparse_major_coder, sparse_major_decoder,
     XHDR_PROTECTED, false },
